/*  Gtk+ Widget Factory
 *  Copyright (C) 1999-2000  Jeroen Benckhuijsen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "gwf_settings.h"

typedef struct _GwfSettingsForeach	GwfSettingsForeach;

struct _GwfSettingsForeach
{
  GwfSettingsForeachFunc	func;
  gpointer 			user_data;
};

static gboolean		gwf_settings_free_foreach 	(gpointer key,gpointer value,gpointer user_data);
static void		foreach_setting			(gpointer key,gpointer value,gpointer user_data);

GwfSettingsList*
gwf_settingslist_new (void)
{
  GwfSettingsList	*list;
  
  list = g_malloc(sizeof(GwfSettingsList));
  list->settings = g_hash_table_new(g_str_hash,g_str_equal);
  
  return list;
}

void
gwf_settingslist_free (GwfSettingsList *list)
{
  g_return_if_fail(list != NULL);
  
  g_hash_table_foreach_remove(list->settings,gwf_settings_free_foreach,NULL);
  g_hash_table_destroy(list->settings);
  g_free(list);
}

gint
gwf_settingslist_add (GwfSettingsList *list,GwfSettings *settings)
{
  g_return_val_if_fail(list != NULL,0);
  g_return_val_if_fail(settings != NULL,0);
  g_return_val_if_fail(settings->name != NULL,0);
  
  if (g_hash_table_lookup(list->settings,(gconstpointer)settings->name))
    return 0;
    
  g_hash_table_insert(list->settings,(gpointer)settings->name,(gpointer)settings);
  return 1;
}

void
gwf_settingslist_remove	(GwfSettingsList *list,gchar *name)
{
  g_return_if_fail(list != NULL);
  g_return_if_fail(name != NULL);
  
  g_hash_table_remove(list->settings,(gconstpointer)name);
}

GwfSettings*
gwf_settingslist_get (GwfSettingsList *list,gchar *name)
{
  g_return_val_if_fail(list != NULL,NULL);
  g_return_val_if_fail(name != NULL,NULL);
  
  return (GwfSettings*)g_hash_table_lookup(list->settings,(gconstpointer)name);
}

void
gwf_settingslist_foreach (GwfSettingsList *list,GwfSettingsForeachFunc func,gpointer user_data)
{
  GwfSettingsForeach	tmp;
  g_return_if_fail(list != NULL);
  g_return_if_fail(func != NULL);
  
  tmp.func = func;
  tmp.user_data = user_data;

  g_hash_table_foreach(list->settings,foreach_setting,(gpointer)&tmp);
}

GwfSettings*
gwf_settings_new (gchar *name)
{
  GwfSettings	*settings;
  
  settings = g_malloc(sizeof(GwfSettings));
  
  settings->name = g_strdup(name);
  settings->create = NULL;
  
  return settings;
}

void
gwf_settings_free (GwfSettings *settings)
{
  g_return_if_fail(settings != NULL);
  
  g_free(settings->name);
  g_free(settings);
}


void
gwf_settings_set_create	(GwfSettings *settings,GwfSettingsCreate func)
{
  g_return_if_fail(settings != NULL);

  settings->create = func;
}

GtkWidget*
gwf_settings_create (GwfSettings *settings,GwfProjlist *projlist)
{
  g_return_val_if_fail(settings != NULL,NULL);
  g_return_val_if_fail(settings->create != NULL,NULL);
  
  return settings->create(projlist);
}


static gboolean
gwf_settings_free_foreach (gpointer key,gpointer value,gpointer user_data)
{
  gwf_settings_free((GwfSettings*)value);

  return TRUE;
}

static void
foreach_setting(gpointer key,gpointer value,gpointer user_data)
{
  GwfSettingsForeach	*tmp = user_data;
  
  tmp->func((GwfSettings*)value,tmp->user_data);
}
