/*  Gtk+ Widget Factory
 *  Copyright (C) 1999-2000  Jeroen Benckhuijsen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gmodule.h>
#include <gtk/gtk.h>
#include <gwf_lib.h>

#include "gwf_projtree_signal.h"
#include "gwf_source_signal.h"
#include "gwf_widprop_signal.h"
#include "gwf_xmlio_signal.h"
#include "gwf_projdata_signal.h"

#define	N_ERR		4

static GwfErrData		errmsg[N_ERR] =	   {
							  {"XmlIO : Signal no name",GWF_ERROR,"Signal encountered without a name"},
							  {"XmlIO : Param no name",GWF_ERROR,"Parameter encountered without a name in signal %s"},
							  {"XmlIO : Duplicate param name",GWF_ERROR,"Two parameters with the same name encountered"},
							  {"Source : Signal : Invalid signal",GWF_ERROR,"One or more of the signal definitions are invalid"}
 						   };
 						
G_MODULE_EXPORT GwfPlugin*
gwf_new_signal_plugin (void)
{
  GwfPlugin	*tmp;
  gint		i;

  tmp = gwf_plugin_new("Signals");
  tmp->version = g_strdup("0.1.0");
  tmp->href = g_strdup("http://gwf.sourceforge.net/Signals");

  tmp->load_plugin = gwf_xmlio_load_ofrsig;
  tmp->save_plugin = gwf_xmlio_save_ofrsig;

  tmp->new_data = gwf_project_signal_new;

  tmp->setting_create = gwf_widpropsig_new;

  tmp->source_check = gwf_source_signal_check;
  gwf_plugin_add_srcwrt(tmp,TYPE,gwf_source_signal_write_signal_defs,HFILE);
  gwf_plugin_add_srcwrt(tmp,TYPEDEF,gwf_source_signal_write_marsh_tdef,CFILE);
  gwf_plugin_add_srcwrt(tmp,TYPE,gwf_source_signal_write_enum,CFILE);
  gwf_plugin_add_srcwrt(tmp,FUNC_DECL,gwf_source_signal_write_marshaller_def,CFILE);
  gwf_plugin_add_srcwrt(tmp,VARIABEL,gwf_source_signal_write_sigarray,CFILE);
  gwf_plugin_add_srcwrt(tmp,FUNCTION,gwf_source_signal_write_classinit,CFILE);
  gwf_plugin_add_srcwrt(tmp,FUNCTION,gwf_source_signal_write_marshaller,CFILE);

  tmp->tree_attach = gwf_projtree_signal_attach;
  tmp->tree_free = NULL;

  for (i = 0 ; i < N_ERR; i++)
  {
    g_ptr_array_add(tmp->errmsg,&errmsg[i]);
  }

  return tmp;
}
