/*  Gtk+ Widget Factory
 *  Copyright (C) 1999-2000  Jeroen Benckhuijsen
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <gwf_lib.h>
#include "gwf_source_signal.h"
#include "gwf_projdata_signal.h"
#include <gtk/gtk.h>
#include <string.h>

static void		check_foreach_signal		(offersig *signal,gpointer user_data);
static void             check_foreach_param		(sigparam *param,gpointer user_data);
static void		write_h_foreach_signal		(offersig *signal,gpointer user_data);
static void		write_h_foreach_param		(sigparam *param,gpointer user_data);
static void		write_marsh_tdef_foreach_signal	(offersig *signal,gpointer user_data);
static void		write_enum_foreach_signal	(offersig *signal,gpointer user_data);
static void		write_marsh_def_foreach_signal	(offersig *signal,gpointer user_data);
static void		write_signal_array_foreach_signal(offersig *signal,gpointer user_data);
static void		write_signal_init_foreach_signal(offersig *signal,gpointer user_data);
static void		write_marshaller_foreach_signal	(offersig *signal,gpointer user_data);
static gchar*		get_marshaller_name		(offersig *signal);
static gchar*		get_marshaller_tdef		(offersig *signal);
static gint		check_marshaller_present	(offersig *signal);

static gchar		*std_marsh[] = {
					"BOOL:NONE",
					"BOOL:POINTER",
					"BOOL:POINTER,POINTER,INT,INT",
					"BOOL:POINTER,INT,INT",
					"BOOL:POINTER,INT,INT,UINT",
					"BOOL:POINTER,STRING,STRING,POINTER",
					"ENUM:ENUM",
					"INT:POINTER",
					"INT:POINTER,CHAR,CHAR",
					"NONE:BOOL",
					"NONE:BOXED",
					"NONE:C_CALLBACK",
					"NONE:C_CALLBACK,C_CALLBACK",
					"NONE:ENUM",
					"NONE:ENUM,FLOAT",
					"NONE:ENUM,FLOAT,BOOL",
					"NONE:INT",
					"NONE:INT,INT",
					"NONE:INT,INT,POINTER",
					"NONE:NONE",
					"NONE:OBJECT",
					"NONE:POINTER",
					"NONE:POINTER,INT",
					"NONE:POINTER,POINTER",
					"NONE:POINTER,POINTER,POINTER",
					"NONE:POINTER,STRING,STRING",
					"NONE:POINTER,UINT",
					"NONE:POINTER,UINT,ENUM",
					"NONE:POINTER,POINTER,UINT,UINT",
					"NONE:POINTER,INT,INT,POINTER,UINT,UINT",
					"NONE:POINTER,UINT,UINT",
					"NONE:POINTER,UINT,UINT",
					"NONE:STRING",
					"NONE:STRING,INT,POINTER",
					"NONE:UINT",
					"NONE:UINT,POINTER,UINT,ENUM,ENUM,POINTER",
					"NONE:UINT,POINTER,UINT,UINT,ENUM",
					"NONE:UINT,STRING"
				       };

#define			N_STD_MARSH	38
				
gint
gwf_source_signal_check	(GwfProject *project,GwfError *error)
{
  gint		result = 1;

  g_return_val_if_fail(project != NULL,0);
  g_return_val_if_fail(error != NULL,0);

  gwf_project_ofrsig_foreach(project,check_foreach_signal,(gpointer)&result);

  if (!result)
    gwf_error_raise_error(error,"Source : Signal : Invalid signal");

  return result;
}


void
gwf_source_signal_write_signal_defs (GwfProject *project,GString *src)
{
  gint		pos;
  GwfSource	srcwrt;

  g_return_if_fail(project != NULL);
  g_return_if_fail(src != NULL);

  pos = gwf_source_find_text(src,"parent_class;");
  g_return_if_fail (pos != -1);

  pos+= 14; /* Skip first element (parent_class) */
  pos = gwf_source_insert(src,pos,"\n");

  srcwrt.project = project;
  srcwrt.src = src;
  srcwrt.result = pos;

  gwf_project_ofrsig_foreach(project,write_h_foreach_signal,(gpointer)&srcwrt);
}

void
gwf_source_signal_write_marsh_tdef (GwfProject *project,GString *src)
{
  g_return_if_fail(project != NULL);
  g_return_if_fail(src != NULL);

  gwf_project_ofrsig_foreach(project,write_marsh_tdef_foreach_signal,(gpointer)src);
}

void
gwf_source_signal_write_enum (GwfProject *project,GString *src)
{
  g_return_if_fail(project != NULL);
  g_return_if_fail(src != NULL);

  if (gwf_project_ofrsig_count(project) > 0)
  {
    g_string_append(src,"\nenum {\n");
    gwf_project_ofrsig_foreach(project,write_enum_foreach_signal,(gpointer)src);
    g_string_append(src,"\n  LAST_SIGNAL\n};\n\n");
  }
}

void
gwf_source_signal_write_marshaller_def(GwfProject *project,GString *src)
{
  g_return_if_fail(project != NULL);
  g_return_if_fail(src != NULL);

  gwf_project_ofrsig_foreach(project,write_marsh_def_foreach_signal,(gpointer)src);
}

void
gwf_source_signal_write_sigarray(GwfProject *project,GString *src)
{
  g_return_if_fail(project != NULL);
  g_return_if_fail(src != NULL);

  if (gwf_project_ofrsig_count(project) > 0)
    g_string_append(src,"\nstatic guint signals[LAST_SIGNAL] = { 0 };\n\n");
}

void
gwf_source_signal_write_classinit(GwfProject *project,GString *src)
{
  gint		pos;
  GwfSource	srcwrt;

  g_return_if_fail(project != NULL);
  g_return_if_fail(src != NULL);

  pos = gwf_source_find_text(src,"parent_class = gtk_type_class");
  g_return_if_fail(pos != -1);
  pos = gwf_source_find_text_from_pos(src,pos,";");
  pos+=2;

  srcwrt.result = pos;
  srcwrt.src = src;
  srcwrt.project = project;

  if (gwf_project_ofrsig_count(project) > 0)
  {
    srcwrt.result = gwf_source_insert(src,srcwrt.result,"\n");
    gwf_project_ofrsig_foreach(project,write_signal_array_foreach_signal,(gpointer)&srcwrt);
    srcwrt.result = gwf_source_insert(src,srcwrt.result,"\n  gtk_object_class_add_signals (object_class, signals, LAST_SIGNAL);\n\n");
    gwf_project_ofrsig_foreach(project,write_signal_init_foreach_signal,(gpointer)&srcwrt);
  }
}

void
gwf_source_signal_write_marshaller(GwfProject *project,GString *src)
{
  g_return_if_fail(project != NULL);
  g_return_if_fail(src != NULL);

  gwf_project_ofrsig_foreach(project,write_marshaller_foreach_signal,(gpointer)src);
}

static void
check_foreach_signal(offersig *signal,gpointer user_data)
{
  gint		*result = user_data;

  if (!signal->name) *result = 0;
  if (!signal->retval) *result = 0;
  if (signal->rvtype == GTK_TYPE_INVALID) *result = 0;
  gwf_ofrsig_param_foreach(signal,check_foreach_param,(gpointer)result);
}

static void
check_foreach_param(sigparam *param,gpointer user_data)
{
  gint		*result = user_data;

  if (!param->name) *result = 0;
  if (!param->type) *result = 0;
  if (param->gtktype == GTK_TYPE_INVALID) *result = 0;
}

static void
write_h_foreach_signal(offersig *signal,gpointer user_data)
{
  GwfSource	*srcwrt = user_data;
  GString	*src = srcwrt->src;
  gchar		*structname;
  gchar		*def1,*def2;
  gint		pos = srcwrt->result;

  pos = gwf_source_insert_tabs_after_printf(src,pos,11,"  %s",signal->retval);
  pos = gwf_source_insert_tabs_after_printf(src,pos,25,"(*%s)",signal->name);

  structname = gwf_source_get_widname_as_struct(srcwrt->project);
  gwf_source_get_widname_as_def(srcwrt->project,&def1,&def2);
  if (def2)
  {
    g_free(def1);
    def1 = def2;
  }

  g_strdown(def1);
  pos = gwf_source_insert_printf(src,pos,"(%s *%s",structname,def1);
  g_free(def1);

  srcwrt->result = pos;

  gwf_ofrsig_param_foreach(signal,write_h_foreach_param,(gpointer)srcwrt);
  srcwrt->result = gwf_source_insert(src,srcwrt->result,");\n");
}

static void
write_h_foreach_param(sigparam *param,gpointer user_data)
{
  GwfSource	*srcwrt = user_data;
  GString	*src = srcwrt->src;

  srcwrt->result = gwf_source_insert_printf(src,srcwrt->result,",%s %s",param->type,param->name);
}

static void
write_marsh_tdef_foreach_signal(offersig *signal,gpointer user_data)
{
  GString	*src = user_data;
  gchar		*marshdef;
  GList		*param;

  if (check_marshaller_present(signal))
    return;

  marshdef = get_marshaller_tdef(signal);

  g_string_sprintfa(src,"\ntypedef %s (*%s) (\n\t\t",signal->retval,marshdef);
  g_string_append(src,"GtkObject*");

  for (param = signal->param ; param ; param = param->next)
    g_string_sprintfa(src,", %s",((sigparam*)param->data)->type);

  g_string_append(src,", gpointer\n);\n\n");

  g_free(marshdef);
}

static void
write_enum_foreach_signal(offersig *signal,gpointer user_data)
{
  GString	*src = user_data;
  gchar		*tmp;

  tmp = g_strdup(signal->name);
  g_strup(tmp);
  g_string_sprintfa(src,"  %s,\n",tmp);
  g_free(tmp);
}

static void
write_marsh_def_foreach_signal(offersig *signal,gpointer user_data)
{
  GString	*src = user_data;
  gchar		*marshaller;

  if (check_marshaller_present(signal))
    return;

  gwf_source_add_tabs_after(src,"static void",8);
  marshaller = get_marshaller_name(signal);
  g_string_sprintfa(src,"%s(GtkObject *object,GtkSignalFunc func,gpointer func_data,GtkArg *args);\n\n",marshaller);

  g_free(marshaller);
}

static void
write_signal_array_foreach_signal(offersig *signal,gpointer user_data)
{
  GwfSource	*srcwrt = user_data;
  gchar		*tmp,*structname,*marshaller;
  gint		nparams;
  GList		*params;

  tmp = g_strdup(signal->name);
  g_strup(tmp);

  structname = gwf_source_get_widname_as_struct(srcwrt->project);
  marshaller = get_marshaller_name(signal);

  for (nparams = 0,params = signal->param ; params ; params = params->next,nparams++);

  srcwrt->result = gwf_source_insert_printf(srcwrt->src,srcwrt->result,"  signals[%s] = \n",tmp);
  srcwrt->result = gwf_source_insert_printf(srcwrt->src,srcwrt->result,"    gtk_signal_new (\"%s\",\n",signal->name);
  srcwrt->result = gwf_source_insert_printf(srcwrt->src,srcwrt->result,"                    GTK_RUN_FIRST,\n");
  srcwrt->result = gwf_source_insert_printf(srcwrt->src,srcwrt->result,"                    object_class->type,\n");
  srcwrt->result = gwf_source_insert_printf(srcwrt->src,srcwrt->result,"                    GTK_SIGNAL_OFFSET (%sClass, %s),\n",structname,signal->name);
  srcwrt->result = gwf_source_insert_printf(srcwrt->src,srcwrt->result,"                    %s,\n",marshaller);
  srcwrt->result = gwf_source_insert_printf(srcwrt->src,srcwrt->result,"                    %s,%d",gtk_type_to_string(signal->rvtype),nparams);

  if (nparams != 0)
  {
    srcwrt->result = gwf_source_insert(srcwrt->src,srcwrt->result,",\n");
    for (params = signal->param ; params ; params = params->next)
    {
      srcwrt->result = gwf_source_insert_printf(srcwrt->src,srcwrt->result,"                    %s",gtk_type_to_string(((sigparam*)params->data)->gtktype));
      if (params->next)
        srcwrt->result = gwf_source_insert(srcwrt->src,srcwrt->result,",\n");
    }
  }
  srcwrt->result = gwf_source_insert(srcwrt->src,srcwrt->result,");\n");

  g_free(tmp);
  g_free(structname);
  g_free(marshaller);
}

static void
write_signal_init_foreach_signal(offersig *signal,gpointer user_data)
{
  GwfSource	*srcwrt = user_data;

  srcwrt->result = gwf_source_insert_printf(srcwrt->src,srcwrt->result,"  klass->%s = NULL;\n",signal->name);
}

static void
write_marshaller_foreach_signal(offersig *signal,gpointer user_data)
{
  GString	*src = user_data;
  gchar		*marshaller;
  gchar		*marshdef;
  gint		nparams = 0,i = 0;
  GList		*param;

  if (check_marshaller_present(signal))
    return;

  g_string_append(src,"static void\n");
  marshaller = get_marshaller_name(signal);
  marshdef = get_marshaller_tdef(signal);
  g_string_sprintfa(src,"%s(GtkObject *object,GtkSignalFunc func,gpointer func_data,GtkArg *args)\n{\n",marshaller);
  g_string_sprintfa(src,"  %s rfunc;\n",marshdef);
  if (signal->rvtype != GTK_TYPE_NONE)
    g_string_sprintfa(src,"  %s\t*return_val\n",signal->retval);
  g_string_sprintfa(src,"  rfunc = (%s)func;\n",marshdef);

  if (signal->rvtype != GTK_TYPE_NONE)
  {
    for (param = signal->param ; param ; param = param->next)
      nparams++;

    g_string_sprintfa(src,"  return_val = GTK_RETLOC_%s(args[%d]);\n",gtk_type_to_string(signal->rvtype) + 9,nparams);
    g_string_append(src,"  *return_val =");
  }

  g_string_append(src,"  (*rfunc)(object");
  for (param = signal->param ; param ; param = param->next,i++)
    g_string_sprintfa(src,", GTK_VALUE_%s(args[%d])",gtk_type_to_string(((sigparam*)param->data)->gtktype) + 9,i);
  g_string_append(src,", func_data);\n");

  g_string_append(src,"}\n\n");

  g_free(marshaller);
  g_free(marshdef);
}

static gchar*
get_marshaller_name (offersig *signal)
{
  GString	*tmp;
  gchar		*marsh;
  GList		*param;

  tmp = g_string_new(NULL);

  g_string_sprintfa(tmp,"gtk_marshal_%s_",gtk_type_to_string(signal->rvtype) + 9);

  if (!signal->param)
    g_string_append(tmp,"_NONE");
  else
    for (param = signal->param ; param ; param = param->next)
      g_string_sprintfa(tmp,"_%s",gtk_type_to_string( ((sigparam*)param->data)->gtktype) + 9);

  marsh = tmp->str;
  g_string_free(tmp,FALSE);

  return marsh;
}

static gchar*
get_marshaller_tdef (offersig *signal)
{
  GString	*tmp;
  gchar		*marsh;
  GList		*param;

  tmp = g_string_new(NULL);

  g_string_sprintfa(tmp,"GtkSignal_%s_",gtk_type_to_string(signal->rvtype) + 9);

  if (!signal->param)
    g_string_append(tmp,"_NONE");
  else
    for (param = signal->param ; param ; param = param->next)
      g_string_sprintfa(tmp,"_%s",gtk_type_to_string( ((sigparam*)param->data)->gtktype) + 9);

  marsh = tmp->str;
  g_string_free(tmp,FALSE);

  return marsh;
}

static gint
check_marshaller_present (offersig *signal)
{
  GList		*marsh = NULL;
  GList		*param = NULL;
  GList		*tmp_marsh,*tmp2;
  gint		i;
  gchar		*tmp;
  gint		marsh_idx = 0;

  for ( i = 0 ; i < N_STD_MARSH ; i++)
    marsh = g_list_append(marsh,GINT_TO_POINTER(i));

  /* Check return value */
  tmp = gtk_type_to_string(signal->rvtype) + 9;

  for (i = 0 ; i < N_STD_MARSH ; i++)
    if (strncmp(tmp,std_marsh[i] + marsh_idx,strlen(tmp)) != 0)
      marsh = g_list_remove(marsh,GINT_TO_POINTER(i));

  /* Check parameters */
  marsh_idx += strlen(tmp);

  if (!signal->param)
  {
    for (i = 0 ; i < N_STD_MARSH ; i++)
    {
      if (strcmp(std_marsh[i] + marsh_idx, ":NONE") != 0)
        marsh = g_list_remove(marsh,GINT_TO_POINTER(i));
    }
    return (marsh) ? 1 : 0;
  }

  for (param = signal->param; marsh && param ; param = param->next)
  {
    tmp = gtk_type_to_string(((sigparam*)param->data)->gtktype) + 9;

    for (tmp2 = tmp_marsh = g_list_copy(marsh) ; tmp_marsh ; tmp_marsh = tmp_marsh->next)
    {
      i = GPOINTER_TO_INT(tmp_marsh->data);
      if ((*(std_marsh[i] + marsh_idx) == '\0') || (strncmp(tmp,std_marsh[i] + marsh_idx + 1,strlen(tmp)) != 0))
        marsh = g_list_remove(marsh,GINT_TO_POINTER(i));
    }
    g_list_free(tmp2);

    marsh_idx += strlen(tmp) + 1;
  }

  /* Check wheter there are no excessive parameters */
  for ( tmp2 = tmp_marsh = g_list_copy(marsh) ; tmp_marsh && marsh ; tmp_marsh = tmp_marsh->next)
  {
    if (*(std_marsh[GPOINTER_TO_INT(tmp_marsh->data)] + marsh_idx) != '\0')
      marsh = g_list_remove(marsh,tmp_marsh->data);
  }

  g_list_free(tmp2);

  return (marsh) ? 1 : 0;
}
